package DataCash::External;

use CCCheck;
use Config;
use Crypt::Blowfish;
use Data::Dumper;
use Exporter;
use FreezeThaw qw(freeze thaw);
use Carp;
use IO::Socket;

@ISA = qw (DataCash);
$VERSION="2.9.0";

sub new {

   my($class,$initialiser) = @_;
   my $self = {} ; 
   bless $self, ref($class) || $class;
   $self = $self->_initialiser($initialiser);
   return $self;
}

sub _initialiser {
    my ($self) = shift;
    my ($fields, $defaults) = "";
    $self->{store}->{currency} = "GBP"; # default value
    $self->{store}->{verno} = "DataCash Perl $VERSION";
    if (@_) {
      $self->client(shift);
    }
    $self->{defaults}->{back_addr} = "new-auth.datacash.com";
    $self->{defaults}->{back_sock} = "8001";
    $self->{defaults}->{timeout}   = "60";
    $self->protocol('0');
    return $self;
}

sub _frontd_message {

	my $msg = shift;
	my $res = '';
	$res .= "ACCOUNT ".$msg->account."\n" if ( $msg->account );
	$res .= "SUM ".$msg->amount."\n" if ( $msg->amount );
	$res .= "AUTH ".$msg->auth."\n" if ( $msg->auth );
	$res .= "CHEQUE ".$msg->cheque."\n" if ( $msg->cheque );
	$res .= "CHTYPE ".$msg->chtype."\n" if ( $msg->chtype );
	$res .= "CLIENT ".$msg->client."\n" if ( $msg->client );
	$res .= "PASS ".$msg->client_pass."\n" if ( $msg->client_pass );
	$res .= "CURRENCY ".$msg->currency."\n" if ( $msg->currency );
	$res .= "CV2 ".$msg->cv2."\n" if ( $msg->cv2 );
	$res .= "EXP ".$msg->exp."\n" if ( $msg->exp );
	$res .= "ISSUE ".$msg->issue."\n" if ( $msg->issue );
	$res .= "NUM ".$msg->num."\n" if ( $msg->num );
	$res .= "POLICY ".$msg->policy."\n" if ( $msg->policy );
	$res .= "POSTCODE ".$msg->postcode."\n" if ( $msg->postcode );
	$res .= "PROTOCOL ".$msg->protocol."\n" if ( $msg->protocol );
	$res .= "REF ".$msg->ref."\n" if ( $msg->ref );
	$res .= "SORTCODE ".$msg->sortcode."\n" if ( $msg->sortcode );
	$res .= "STREET_ADDRESS ".$msg->street_address."\n" if ( $msg->street_address );
	$res .= "SWIPE ".$msg->swipe."\n" if ( $msg->swipe );
	$res .= "TYPE ".$msg->type."\n" if ($msg->type );
	$res .= "VERNO ".$msg->verno."\n" if ( $msg->verno );
	$res .= ".\n";

	return $res;
}

sub send {

   my ($self) = shift;
   my ($class) = ref ($self) || croak "self is not an object\n";
   my ($ck) = $self->final_sanity_check();
   if ($ck !~ /^0/) {
     ($self->return_code() eq 'negative') ? return ('-'.$ck) : return $ck;
  }
   my $tmpnum = $self->num();

   if ((defined($DataCash::Blowfish)) && (length($tmpnum) < 20)) {
       $self->num($self->big_encrypt($self->num()));
   }
   my ($ref) = $self->ref();

   if (($ref eq "") ||
       (length($ref) < 6) ||
       (length($ref) > 16)) {
       return '22';
   }

  # Let's try to send this to the frontd
  $response = _send_to_server(  $self->_frontd_message, $self->{defaults}->{back_addr}, $self->{defaults}->{back_sock} );

  return $response;
}

sub _send_to_server {
  my ( $msg, $host, $socket ) = @_;
  
#  print STDERR "Sending to host $host socket $socket\n";
  
  my $sock = new IO::Socket::INET (
				   PeerAddr => $host,
				   PeerPort => $socket,
				   Proto     => "tcp");
  
  die "Failed to connect to $host:$socket: $@" unless ( defined $sock );
  
#  print STDERR "Local Port: ", $sock->sockport, "\n";
  $sock->autoflush(1);
  
  my $response = "";
  if ($Config{osname} =~ /mswin/i) { # windoze doesn't grok alarm( )
    $sock->print($msg);
    # $sock->shutdown( 1 );
    shutdown $sock, 1;
    $response = $sock->getline();
  } else {
    $SIG{ALRM} = sub {
      die "timeout";
    };
    eval {
      alarm(60);
      $sock->print($msg);
      # $sock->shutdown( 1 );
      shutdown $sock, 1;
      $response = $sock->getline();
      alarm(0);
    };
    
    if ($@ =~ /timeout/i) {
      close( $sock );
      print STDERR "timeout trap: $@\n";
      return "3:timeout waiting for frontd to reply\n";
    } 
  }
  
  return $response;
}

sub final_sanity_check {
  
  my $self = shift;
  if ($self->type() =~ /fulfill/i) { return 0; }

  if ($self->type() =~ /cheque/i) {
    if ($self->{store}->{sortcode} =~ /\D/) {
      $self->{store}->{sortcode} =~ s/\D+//g;
    }
    my %ch_err = (
		  1 => "bad cheque number",
		  2 => "bad sortcode",
		  3 => "bad cheque number and sortcode",
		  4 => "bad account number",
		  5 => "bad account anc cheque number",
		  6 => "bad account and sortcode",
		  7 => "bad cheque, sortcode and account"
		 );
    my ($bad_cheque) = 1;
    my ($bad_sortcode) = 2;
    my ($bad_account) = 4;
    my $err = 0;
    if (length($self->cheque()) != 6) {
      $err += $bad_cheque;
    }
    if (length($self->{store}->{sortcode}) != 6) {
      $err += $bad_sortcode;
    }
    if ((length($self->{store}->{account}) > 10) or 
	(length($self->{store}->{account}) == 0)){
      $err += $bad_account;
    }
    if ($err > 0) {
      my $r = 40 + $err;
      my $rstr = "$r:";
      $rstr .= $ch_err{$err};
      $rstr .= ":::";
      return $rstr;
    } else {
      return "0:";
    }
  } 
  else {
    return "0:" if (length($self->num()) > 20);
    my $cc = CCCheck->new();
    $cc->card_number($self->num());
    $cc->issue_number($self->issue());
    $cc->expiry_date($self->exp());
    if ($cc->check() == 1 )  { return 0 }; # card is OK
    if ($cc->return_code() =~ /^33/) {     # card number encrypted - ignore
      return 0;
    }
    return $cc->return_code();             # return some error code
  }
}

sub big_encrypt {
    my $self = shift @_;
    my $plaintext = shift @_;
    my $cypher = Crypt::Blowfish->new($DataCash::Blowfish);
    my ($section, $cyphertext, $tmp) = '';
    while ($section = substr($plaintext, 0, 8)) {
        $section = pack("a8", $section);
        $tmp = $cypher->encrypt ($section);
        $cyphertext .= unpack("H16", $tmp);
        $plaintext = substr($plaintext, 8);
    }
    return $cyphertext;
} 

sub version {
  my $self = shift;
  return $self->verno;
}

sub hostname {
  my $self = shift;
  @_ ? return $self->{'defaults'}->{'back_addr'} = shift : return $self->{'defaults'}->{'back_addr'};
}

sub socket {
  my $self = shift;
  @_ ? return $self->{'defaults'}->{'back_sock'} = shift : return $self->{'defaults'}->{'back_sock'};
}

sub AUTOLOAD {
  
  my ($self) = shift;
  my ($type) = ref ($self) || croak "$self is not an object\n";
  my ($name) = $AUTOLOAD;
  $name =~ s/.*://;
  $name = lc($name);
  return if ($name =~ /destroy/i);
  if (@_) {
    my (@tmp) = @_;
    # $tmp[0] =~ s/\s+//g; 
    return $self->{'store'}->{$name} = "$tmp[0]";
  } else {
    return $self->{'store'}->{$name};
  }
}


1;

=pod 

=head1 NAME 

DataCash::External -- external encapsulation object

=head1 SYNOPSIS

This is the external Perl 5 interface to the DataCash server system. It is
suitable for integration into Perl applications which need to carry out
credit card transactions and have a real-time internet connection available.

The following description is somewhat technical and presupposes a working
knowledge of Perl.  DataCash::External is not a standalone program for 
carrying out credit card authorisations; it is a toolkit for building such 
programs.  If you are not a Perl programmer, you will need to find one who 
can help you use this package.

=head1 DESCRIPTION

This is a Perl 5 class used to encapsulate data sent by a user who
wants to authorize a credit card transaction. In a former life it was 
called CC::Transaction. 

To submit a transaction to a DataCash 2.I<x> server, you need to build and
populate a DataCash::External object.

To build a DataCash::External object, first initialise it using the
vTID of the terminal for which the transaction is being processed:

  my ($vtid) = "20104001"; # Virtual Terminal ID number assigned by DataCash
  my $obj = DataCash::External->new($vtid);

This creates a DataCash::External object which has various methods for 
collating details, creating a message in the correct format and sending this
message to the DataCash system.

=head2 Mandatory Fields

For an APACS transaction, certain mandatory methods I<must> be used
to populate some fields in the object; otherwise the transaction I<will>
fail. These include:

=over 4

=item num

Credit card number (aka PAN, primary account number, if you speak Banking).
This is encrypted using Blowfish by the merchant, using the merchant\'s
private key. 

=item amount

The amount should be completely specified, i.e. 100 should be 100.00.
Currencies with different numbers of decimal fractions should also be
fully specified with the correct number of digits (eg JOD should have 3
digits after the decimal place). 

=item exp

Expiry date of the card (MM/YY format).

=item issue

I<Either> the issue number of the card (if it has one) or the start date
of the card (MM/YY format) (if it has a start date but no issue number).
(NB: this field is only mandatory for cards that have start dates or
issue numbers.)

=item ref

A B<unique> six-to-twelve digit identifier (except for B<fulfill> requests
- see below).

When submitting a B<auth>, B<refund>, or B<pre> request, identifiers must 
be a unique merchant-assigned string. Uniqueness is checked on a per-vTID
 basis and non-unique IDs I<will> be rejected.

When submitting a B<fulfill> request, the reference should contain the
DataCash-returned transaction reference from the previous B<pre> request 
which is being fulfilled. This will typically be 16 digits long.

=item client

The Virtual Terminal ID (vTID) as assigned by DataCash. This is required
as additional confirmation that the vTID is correct when processing a
transaction.

=item client_pass

The password for this terminal as assigned by DataCash. (Yes, it's another
authentication check.)

=item type

Can accept one of the following values:

=over 6

=item auth

This is an authorisation request.  If successful, the customer's credit
card will be debited for the specified amount, and the transaction will
be released for clearing via the acquiring bank.

=item refund

This is a refund request. If successful, the customer's credit card
will be credited with the specified amount, and the transaction will
be cleared via the acquiring bank. (The money comes out of the I<merchant's>
account.)

=item pre

Pre-authorise a transaction. A credit check is carried out, and the
requested amount is reserved on the customer's credit account. Returns an
authorisation code and a unique reference number which is typically 
16 digits long. 

The transaction is I<not> released for clearing via the acquiring bank until
a subsequent B<fulfill> request is sent. If no fulfill request is sent
within a period specified by the bank (typically 5-10 working days) the
pre-auth times out.

=item fulfill

Fulfill a previously pre-authed transaction. You must call the B<auth>
method to specify the authorisation code of the pre-auth you are
fulfilling, and supply the unique transaction reference returned by
the pre-auth as the B<ref> method parameter.

=back

=back

=head2 Additional Fields

Additional method are provided which may be specified depending on the 
type of transaction:

=over 4

=item currency

If not specified, defaults to GBP (pounds sterling). If specified, it
should contain an ISO currency code, and you must be using a vTID that
has been set up to clear transactions in that currency. (Use this method
when submitting foreign currency transactions for settlement.)

=item hostname

If not specified, defaults to new_auth.datacash.com which is where all
live transactions should be sent.

=item socket

If not specified, defaults to 8001 which is where all live transactions
should be sent by this API.

=back

=head2 Local card checking methods

To verify that a credit card number is correct, DataCash::External
supplies a couple of useful methods: these are loaded via B<CCCheck.pm>.

B<final_sanity_check()> basically executes these tests and returns an
appropriate error.

=head2 Responses from the DataCash servers

When you have a successful transaction, you will receive a seven-part
response (delimited by colons) like this:

 1:024543:907927010:2185999900000691:VISA:Barclays Bank Plc:United Kingdom

The first element is the DataCash server response code; 1 indicates
success.  A full list of the response codes are available on the 
DataCash web site (www.datacash.com).

The second element is an I<authorisation code> returned by the bank. Cut
this out and keep it; it's important.

The third element is a UNIX timestamp indicating the time at which the
transaction was processed, to the nearest second. (It's a signed integer
and if you know how to decode UNIX time values you can use it.)

The fourth element is a unique I<internal> DataCash transaction identifier.
This is also important.

The fifth, sixth and seventh elements are respectively the card type,
the issuing bank and the country of issue of the card. These are supplied
for information only, and may be missing if we cannot find bank
information for this card (e.g. if that card range has been added since
we last updated our list of banks).

=head2 Types of transaction

=over 4

=item auth

The simplest type of transaction is an B<auth> request:

   $obj = DataCash::External->new("21859999");
   $obj->client("21859999");
   $obj->client_pass("fred");
   $obj->type("auth");
   $obj->num("5473000000000007");
   $obj->amount("10.01");
   $obj->currency("GBP");
   $obj->exp("11/05");
   $obj->ref("1234567890");
   $result = $obj->send();

An B<auth> request attempts to debit the designated credit card
for an amount of money. If successful, an authorisation is obtained and
the transaction will be cleared into the merchant's account as soon
as the next batch of transactions is submitted to the acquiring bank.
(Batch submission is carried out by DataCash and is not a feature of
this module.)

If a card has an issue number, specify it like this:

   $obj->issue("01");

If a card has a start date, specify it like this:

   $obj->issue("09/99");

(Note that this syntax may be changed to something more sensible at a later 
date.)

=item refund

A B<refund> works exactly the same way, albeit in reverse. Money comes
out of the merchant's account and goes to a customer the next time
transactions are submitted to the acquirer.

   $obj = DataCash::External->new("21859999");
   $obj->client("21859999");
   $obj->type("refund");

=item pre / fulfill

I<Two-phase fulfillment> is a more complex operation. DataCash allows you
to pre-authorise a transaction -- getting permission from the card issuer
to debit a card, and reserving credit on the cardholder's account -- then
release the transaction for clearing at a later date by indicating that
the customer's order has been fulfilled. (If the authorisation code
returned by the pre-auth step is not used within a week or two it times
out, and the customer gets the reserved credit on their card back.)

First, issue a pre-auth request:

   $obj = DataCash::External->new("21859999");
   $obj->client("21859999");
   $obj->type("pre");

It is vital at this stage to record the authorisation code and 
transaction reference, as without these details you will be unable to 
fulfill the transaction at a later date.  These details are contained
within the response string that you receive with each successful 
transaction.

To indicate that you have fulfilled a transaction, prepare a request
that looks like this:

   $obj = DataCash::External->new("21859999");
   $obj->client("21859999");
   $obj->client_pass("fred");
   $obj->type("fulfill");
   # the DataCash internal transaction identifier
   $obj->ref("2185999900000691");
   $obj->auth("024543");     # the authcode returned by the bank
   $obj->amount("10.01");
   $result = $obj->send()

Note that you do I<not> need to resubmit the card details (i.e. card
number, expiry date, issue number or start date).  Also, if the amount
of the B<fulfill> is the same as the amount you originally specified in 
the B<pre> transaction, the fulfill will default to the amount 
original specified in the B<pre> transaction if you do not resubmit 
the amount details.

When you specify an amount to debit the cardholder's account with, 
standard regulations permit a variation of up to +/- 10% (to allow for
variation from the original bill, e.g. restaurant gratuities).

=back

=head2 Sending the Transaction

To send a DataCash::External object to the DataCash back-end server for
processing, do this:

   $result = $obj->send();

That's all it takes!

As noted above, B<$result> then contains a response. This consists of one
to seven fields delimited by colons. The first field is always a DataCash
return code; the remainder consist of a message from the acquiring bank,
and (optionally) a time stamp, a DataCash-assigned transaction reference
and bank details.
A list of DataCash return codes can be found at
http://www.datacash.com/stronghold/returncodes.html

=head2 Encryption 

DataCash::External attempts to encrypt card numbers using Blowfish
(if the Cryptix modules are installed -- see the installation instructions.) 
To encrypt the card number, you must set a public key, supplied by DataCash.
Do this in a begin block at the start of your program, e.g.: 

  BEGIN {
    $DataCash::Blowfish = 'SFGJ9086ADFH086'; # key supplied by DataCash
  }

Thereafter, your credit card number details will be encrypted before being
 despatched to DataCash. 


=head1 METHODS

=head2 new() 

Create a new DataCash::External object. 

Requires, as an initializer, a valid DataCash vTID (8-byte number uniquely
identifying an Apacs EPOS terminal). 

Use the test vTID '21859999' if you don't have a real one yet. 

This method calls an internal method that defines the names of other methods
that DataCash::External knows about (which are used to get or set fields
in the object). 

It returns a reference to the new DataCash::External object. 

=head2 send() 

Despatches a message to the currently configured DataCash server.  The 
configuration details are set when the DataCash::External object is created.
The send() method returns: 

=over 4

=item connection failed 

Unable to communicate with the DataCash system

=item write failed

(bytes written != data to write). this returns additional parameters after
a colon: number of bytes written/bytes to send. 

=item read failed

nothing read

=item unspecified weird error 

Check return codes available on DataCash web site

=item no transaction reference number specified

transmission aborted

=back 

... or a string from the server. (See the documentation at DataCash for
 details of DataCash protocol values.) 

=head2 final_sanity_check() 

Perform final pre-despatch sanity checking on a transaction. Make sure
expiry date (and start date, if present) are correctly formatted and
in bounds. Make sure card number passes LUHN checksum. Make sure the
number is of the correct length and an issue number/start date are
present if required. Return zero for success, or various codes (below)
for errors. 

=head2 AUTOLOAD() 

The AUTOLOAD method is not called explicitly: it simply provides
 virtual methods on demand for fields that the current terminal's
 clearing mechanism needs or knows about. 

=head2 version()

The version method can be called to retrieve the version of API that
you are presently using.

=head2 return_code()

When a transaction is processed, there are many checks done on the data
that has been submitted prior to transmit to our servers.  Once the data
has been received by our servers, even more checks are done on the data 
before an attempt is made to authorise the transaction.

This means that an unsuccessful transaction may not make it as far as our
servers before being rejected.  It is sometimes very useful to know whether
the transaction has been rejected by the API rather than the servers, and
one way to ascertain this is to use the return_code() method.  You use this
in the following way:

 $obj->return_code('negative');

This will therefore mean that any return code that is a negative integer 
will have been returned by the API, while a positive return code will have
been returned by our servers.

Please note that if this method is not called prior to sending the 
transaction, a positive return code will be returned no matter which system 
returned the code.
